import java.io.File;
import java.net.URL;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Iterator;
import java.util.LinkedList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sun.syndication.feed.synd.SyndEntry;
import com.sun.syndication.feed.synd.SyndFeed;
import com.sun.syndication.io.SyndFeedInput;
import com.sun.syndication.io.XmlReader;

/**
 * Basic news collection job, implementing basic configuration handling and collection routines.
 * 
 * @author fhogenboom
 */
public abstract class NewsJob extends CollectJob
{

	protected Integer                 max_hash, max_retry, max_title, mhm_size, num_msgs,
	        zip_buffer, zip_files;
	protected LinkedList<Integer>     mhm;
	protected String                  file_path, file_ext, rss_name, rss_src, txt_cls;

	private static boolean            zipping  = false;
	private static int                mtm_size = 0, num_parts = 0, num_out = 0;
	private static LinkedList<String> mtm      = null;
	private static String             cur_date = null;

	private static Logger             _log     = LoggerFactory.getLogger(NewsJob.class);

	/**
	 * Constructor.
	 */
	public NewsJob()
	{
		super();
	}

	/**
	 * Configures the {@link NewsJob} component by parsing a configuration file for local component
	 * elements.
	 * 
	 * @param job_key ID of the component.
	 * @return {@link Boolean} value indicating whether parsed configuration is valid.
	 */
	protected boolean configureLocal(String job_key)
	{
		try
		{
			// Load news configuration
			max_retry = Util.getInteger(Util.parseXMLelement("news", "max_retry"));
			max_title = Util.getInteger(Util.parseXMLelement("news", "max_title"));
			file_path = Util.getString(Util.parseXMLelement("news", "file_path"));
			file_ext = Util.getString(Util.parseXMLelement("news", "file_ext"));
			zip_files = Util.getInteger(Util.parseXMLelement("news", "zip_files"));
			zip_buffer = Util.getInteger(Util.parseXMLelement("news", "zip_buffer"));

			// Load component configuration
			rss_name = Util.getString(Util.parseXMLid(job_key, "eye", "rss_name"));
			rss_src = Util.getString(Util.parseXMLid(job_key, "eye", "rss_src"));
			txt_cls = Util.getString(Util.parseXMLid(job_key, "eye", "txt_cls"));
			max_hash = Util.getInteger(Util.parseXMLid(job_key, "eye", "max_hash"));
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
		// Return whether parsed configuration is valid
		return (max_retry != null && max_title != null && file_path != null && file_ext != null
		        && zip_files != null && zip_buffer != null && rss_name != null && rss_src != null
		        && txt_cls != null && max_hash != null);
	}

	/**
	 * Configures the {@link NewsJob} component by loading configurations for local component
	 * elements stored in a {@link org.quartz.JobDataMap}, which is available during runtime of the
	 * component.
	 * 
	 * @param context Execution context of the component.
	 * @return {@link Boolean} value indicating whether loaded configuration is valid.
	 */
	@SuppressWarnings("unchecked")
	protected boolean loadLocal(JobExecutionContext context)
	{
		try
		{
			// Load data map of current component
			JobDataMap dataMap = Util.loadDataMap(context);

			// Load news configuration
			max_retry = dataMap.getInt("max_retry");
			max_title = dataMap.getInt("max_title");
			file_path = dataMap.getString("file_path");
			file_ext = dataMap.getString("file_ext");
			zip_files = dataMap.getInt("zip_files");
			zip_buffer = dataMap.getInt("zip_buffer");

			// Load component configuration
			rss_name = dataMap.getString("rss_name");
			rss_src = dataMap.getString("rss_src");
			txt_cls = dataMap.getString("txt_cls");
			max_hash = dataMap.getInt("max_hash");

			// Load data
			try
			{
				mhm_size = dataMap.getInt("mhm_size");
			}
			catch (Exception e)
			{
				mhm_size = 0;
			}
			try
			{
				num_msgs = dataMap.getInt("num_msgs");
			}
			catch (Exception e)
			{
				num_msgs = 0;
			}
			mhm = (LinkedList<Integer>) dataMap.get("mhm");
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
		// Return whether loaded configuration is valid
		return (max_retry != null && max_title != null && file_path != null && file_ext != null
		        && zip_files != null && zip_buffer != null && rss_name != null && rss_src != null
		        && txt_cls != null && max_hash != null);
	}

	/**
	 * Stores the configuration for local component elements of the {@link NewsJob} component in a
	 * {@link org.quartz.JobDataMap}, which is available during runtime of the component.
	 * 
	 * @param context Execution context of the component.
	 */
	protected void storeLocal(JobExecutionContext context)
	{
		try
		{
			// Load data map of current component
			JobDataMap dataMap = Util.loadDataMap(context);

			// Store news configuration
			dataMap.put("max_retry", max_retry);
			dataMap.put("max_title", max_title);
			dataMap.put("file_path", file_path);
			dataMap.put("file_ext", file_ext);
			dataMap.put("zip_files", zip_files);
			dataMap.put("zip_buffer", zip_buffer);

			// Store component configuration
			dataMap.put("rss_name", rss_name);
			dataMap.put("rss_src", rss_src);
			dataMap.put("txt_cls", txt_cls);
			dataMap.put("max_hash", max_hash);

			// Store data
			dataMap.put("mhm_size", mhm_size);
			dataMap.put("num_msgs", num_msgs);
			dataMap.put("mhm", mhm);

			// Store data map of current component
			Util.storeDataMap(context, dataMap);
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
	}

	/**
	 * Collects data from a specific source.
	 */
	protected void collect()
	{
		// Initialize data structures
		if (mtm == null)
		{
			mtm = new LinkedList<String>();
		}
		if (mhm == null)
		{
			getLogger().info("Fetching data from " + rss_src);
			mhm = new LinkedList<Integer>();
		}
		// Zip output files
		zipFiles();
		int count = 0;
		try
		{
			// Open RSS feed
			URL feedSource = new URL(rss_src);
			SyndFeedInput input = new SyndFeedInput();
			SyndFeed feed = input.build(new XmlReader(feedSource));
			Iterator<?> iter = feed.getEntries().iterator();
			// For each entry
			while (iter.hasNext())
			{
				SyndEntry entry = (SyndEntry) iter.next();
				// Get hash
				int hash = entry.hashCode();
				// If hash is not in own memory (of an individual NewsJob)
				if (!inMemory(hash))
				{
					// Add hash to memory
					addToMemory(hash);
					// Get title
					String title = getTitle(entry);
					// If title is not in shared memory (across all NewsJobs)
					if (!inMemory(title))
					{
						// Add title to memory
						addToMemory(title);
						// Save entry to file
						saveToFile(entry);
						// Update counters
						count++;
						num_out++;
						// Zip output files
						zipFiles();
					}
					else
					{
						getLogger().debug("Ignoring message with duplicate title");
					}
				}
				else
				{
					getLogger().debug("Ignoring message with duplicate hash");
				}
			}
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
		// Zip output files
		zipFiles();
		getLogger().info("Collected " + count + " messages from " + rss_name);
	}

	/**
	 * Checks whether a given hash code is stored in the memory of an individual {@link NewsJob}
	 * component.
	 * 
	 * @param hash Hash code of entry to be checked.
	 * @return {@link Boolean} value indicating whether hash is in own memory.
	 */
	private boolean inMemory(int hash)
	{
		return mhm.contains(new Integer(hash));
	}

	/**
	 * Checks whether a given title is stored in the memory of all {@link NewsJob} components.
	 * 
	 * @param title Title of entry to be checked.
	 * @return Boolean value indicating whether title is in shared memory.
	 */
	private boolean inMemory(String title)
	{
		return mtm.contains(title);
	}

	/**
	 * Add a given hash code to the memory of an individual {@link NewsJob} component.
	 * 
	 * @param hash Hash code of entry to be added.
	 */
	private void addToMemory(int hash)
	{
		// If memory is not full
		if (mhm_size < max_hash)
		{
			// Increase counter
			mhm_size++;
		}
		// Else
		else
		{
			// Remove first item from memory
			mhm.remove();
		}
		// Add hash code to memory
		mhm.add(new Integer(hash));
	}

	/**
	 * Add a given title to the memory of all {@link NewsJob} components.
	 * 
	 * @param title Title of entry to be added.
	 */
	private void addToMemory(String title)
	{
		// If memory is not full
		if (mtm_size < max_title)
		{
			// Increase counter
			mtm_size++;
		}
		// Else
		else
		{
			// Remove first item from memory
			mtm.remove();
		}
		// Add title to memory
		mtm.add(title);
	}

	/**
	 * Parses and saves RSS entry to a file using an XML format. File path and extension are
	 * specified in the configuration file. File name is constructed as {@code date}
	 * {@code rss_name} {@code num_msgs}, where {@code date} and {@code num_msgs} represent the
	 * current date and number of messages of an individual {@link NewsJob} component, and
	 * {@code rss_name} represents the name of an RSS feed, specified in the configuration file.
	 * 
	 * @param entry RSS entry to be parsed and saved to file.
	 */
	private void saveToFile(SyndEntry entry)
	{
		// Get data
		String title = getTitle(entry);
		String author = getAuthor();
		String text = getText(entry);
		String publ = getDate();
		String link = getLink(entry);

		// Construct file name
		String file = date + " " + rss_name + " " + num_msgs;
		getLogger().info("Saving news message to file " + file_path + file + file_ext);

		try
		{
			// Create XML document builder
			DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docFactory.newDocumentBuilder();

			// Create XML document with root element
			Document doc = docBuilder.newDocument();
			Element rootElement = doc.createElement("message");
			doc.appendChild(rootElement);

			// Create title element
			Element el_title = doc.createElement("title");
			rootElement.appendChild(el_title);
			el_title.appendChild(doc.createTextNode(title));

			// Create author element
			Element el_author = doc.createElement("author");
			rootElement.appendChild(el_author);
			el_author.appendChild(doc.createTextNode(author));

			// Create date element
			Element el_date = doc.createElement("date");
			rootElement.appendChild(el_date);
			el_date.appendChild(doc.createTextNode(publ));

			// Create text element
			Element el_text = doc.createElement("text");
			rootElement.appendChild(el_text);
			el_text.appendChild(doc.createTextNode(text));

			// Create link element
			Element el_link = doc.createElement("link");
			rootElement.appendChild(el_link);
			el_link.appendChild(doc.createTextNode(link));

			// Transform XML document to DOM document
			TransformerFactory transformerFactory = TransformerFactory.newInstance();
			Transformer transformer = transformerFactory.newTransformer();
			DOMSource source = new DOMSource(doc);

			// In case path does not exist, create it
			new File(file_path).mkdirs();

			// Store XML file
			StreamResult result = new StreamResult(new File(file_path + file + file_ext));
			transformer.transform(source, result);

			// Update counter
			num_msgs++;
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
	}

	/**
	 * Zips output files whenever there is a date swap or the maximum number of files allowed in a
	 * single zip file (specified in the configuration file) is exceeded. File path and extension of
	 * the files to be zipped are specified in the configuration file, as well as the buffer used
	 * for zipping. Zip files are stored in the same directory as the files they contain. After
	 * completion, target files are removed. File name is constructed as {@code cur_date}-
	 * {@code num_parts}, where {@code cur_date} represents the current date (which can be
	 * yesterday's date in case of a date swap) and where {@code num_parts} represents the number
	 * number of zip files that have been created by all {@link NewsJob} components.
	 */
	private void zipFiles()
	{
		// If current date is not specified
		if (cur_date == null)
		{
			// Update date
			cur_date = date;
		}
		// Else
		else
		{
			// If date swap detected or if the maximum number of files allowed in a single zip
			// file is exceeded
			if (!cur_date.equals(date) || num_out >= zip_files)
			{
				// If no NewsJob component is already zipping files
				if (!zipping)
				{
					// Add zip lock
					zipping = true;
					// If there are files to zip
					if (num_out > 0)
					{
						// Zip files
						Util.zipFiles(getClass().getSuperclass().getName(), file_path, cur_date
						        + "-" + ++num_parts, file_ext, zip_buffer, zip_files);
					}
					// If there is a date swap
					if (!cur_date.equals(date))
					{
						// Reset counter
						num_parts = 0;
					}
					// Update files left to be zipped
					num_out -= Math.min(num_out, zip_files);
					// Update date
					cur_date = date;
					// Remove zip lock
					zipping = false;
				}
			}
		}
	}

	/**
	 * Retrieves the author of an RSS entry.
	 * 
	 * @return {@link String} value representing the name of an RSS feed.
	 */
	private String getAuthor()
	{
		return rss_name;
	}

	/**
	 * Retrieves the author of an RSS entry.
	 * 
	 * @param entry RSS entry to be parsed.
	 * @return {@link String} value representing the author of an RSS entry.
	 */
	private String getAuthor(SyndEntry entry)
	{
		return entry.getAuthor();
	}

	/**
	 * Retrieves the publishing date of an RSS entry.
	 * 
	 * @return {@link String} value representing the current date and time.
	 */
	private String getDate()
	{
		return date + " " + time;
	}

	/**
	 * Retrieves the publishing date of an RSS entry.
	 * 
	 * @param entry RSS entry to be parsed.
	 * @return {@link String} value representing the publishing date of an RSS entry.
	 */
	private String getDate(SyndEntry entry)
	{
		DateFormat sdf = new SimpleDateFormat(form_date + " " + form_time);
		return sdf.format(entry.getPublishedDate());
	}

	/**
	 * Retrieves the link to the full text of an RSS entry.
	 * 
	 * @param entry RSS entry to be parsed.
	 * @return {@link String} value representing the link to the full text of an RSS entry.
	 */
	private String getLink(SyndEntry entry)
	{
		return entry.getLink();
	}

	/**
	 * Retrieves the full text of an RSS entry.
	 * 
	 * @param entry RSS entry to be parsed.
	 * @return {@link String} value representing the full text of an RSS entry.
	 */
	private String getText(SyndEntry entry)
	{
		return parseHTML(entry.getLink());
	}

	/**
	 * Retrieves the title of an RSS entry.
	 * 
	 * @param entry RSS entry to be parsed.
	 * @return {@link String} value representing the title of an RSS entry.
	 */
	private String getTitle(SyndEntry entry)
	{
		return entry.getTitle();
	}

	/**
	 * Retrieves the {@link org.slf4j.Logger} of the component.
	 * 
	 * @return {@link org.slf4j.Logger} used for logging system output.
	 */
	protected abstract Logger getLogger();

	/**
	 * Parses the HTML page of a specified link into plain text.
	 * 
	 * @param url Link to be parsed.
	 * @return {@link String} value representing the full text.
	 */
	protected abstract String parseHTML(String url);

}